﻿using System;
using System.Web;
using System.IO;
using System.Text;
using System.Security.Cryptography;
using System.Web.Security;
using System.Security.Principal;

namespace DownloadHandlers
{
    ////////////////////////////////////////////////////
    // Example URLs to call download handler and download a file:
    //
    //    https://localhost/DownloadPortal/Download?file=file.txt
    //    https://localhost/DownloadPortal/Download?file=file.txt&chunksize=1000000
    //    https://localhost/DownloadPortal/Download?file=customer1/file.txt
    //    https://localhost/DownloadPortal/Download?file=customer1/file.txt&chunksize=1000000
    //
    //    Important!!!  Must replace 'localhost' in URL with actual machine name or SSL certificate will fail.
    //
    //    Can either include the 'Range' HTTP request header in the HTTP web request OR
    //    supply the 'chunksize' parameter in the URL. If include the 'Range' HTTP request header,
    //    it will transmit back that portion of the file identified by the 'Range'. If supply the
    //    'chunksize' parameter, it will transmit back the entire file in separate pieces the size of
    //    'chunksize'. If supply both, 'Range' will be used. If supply neither, it will transmit back
    //    the entire file in one piece.
    //
    // Web.config on IIS 7.0:
    //    <system.webServer>
    //      <handlers>
    //        <add name="Download" verb="*" path="Download" type="DownloadHandlers.DownloadHandler" />
    //      </handlers>
    //    </system.webServer>
    //
    // Put DownloadHandler.dll and IOThreads.dll in the /bin directory of the virtual directory on IIS.
    //
    // Put files to be downloaded into virtual directory or sub directory under virtual directory.
    //
    // To debug:
    //    Debug \ Attach to Process \ w3wp.exe
    //
    // Note: Make sure the IIS virtual directory is using an AppPool that is using .NET Framework 4.0.
    //       Define a new AppPool using .NET Framework 4.0, if no other AppPool exists that is using .NET Framework 4.0.
    ////////////////////////////////////////////////////

    public class DownloadHandler : IHttpAsyncHandler
    {
        public IAsyncResult BeginProcessRequest(HttpContext context, AsyncCallback cb, object extraData)
        {
            //Offloads to "completion port threads" inside CLR ThreadPool
            //instead of "worker threads" inside CLR ThreadPool, so that our
            //work doesn't use up the "worker threads" that are needed for the
            //IIS server's request processing. (see comment section at top of
            //IOThread class for more details)
            IOThread.ProcessRequestDelegate d = ProcessRequest;
            return d.IOBeginInvoke(context, cb, extraData);
        }

        public void EndProcessRequest(IAsyncResult result)
        {
            IOThread.IOAsyncResult ar = (IOThread.IOAsyncResult)result;
            IOThread.ProcessRequestDelegate d = (IOThread.ProcessRequestDelegate)ar.AsyncDelegate;
            d.IOEndInvoke(result);
        }

        public void ProcessRequest(HttpContext context)
        {
            try
            {
                string file = context.Request.QueryString["File"];
                if (string.IsNullOrEmpty(file))
                    throw new Exception("Must specify file in query string.  (Example: Download?File=your file)");

                string fileName = Path.GetFileName(file);
                if (string.IsNullOrEmpty(fileName))
                    throw new Exception("File name '" + fileName + "' is not valid.");

                long chunkSize = 0;
                if (!string.IsNullOrEmpty(context.Request.QueryString["ChunkSize"]))
                    if (context.Request.QueryString["ChunkSize"].Trim().Length > 0)
                        chunkSize = long.Parse(context.Request.QueryString["ChunkSize"]);

                FileInfo fi = new FileInfo(context.Server.MapPath(file));
                long fileLength = fi.Length;
                if (chunkSize > 0 && chunkSize > fileLength)
                    throw new Exception("ChunkSize is greater than file length.");


                context.Response.ClearHeaders();
                context.Response.ClearContent();
                context.Response.Clear();

                //request is just checking file length
                if (context.Request.HttpMethod == "HEAD")
                {
                    context.Response.AddHeader("content-length", fileLength.ToString());
                    context.Response.AddHeader("accept-ranges", "bytes");
                    context.Response.Flush();
                    return;
                }
                
                //file save
                context.Response.ContentEncoding = Encoding.UTF8;
                context.Response.AddHeader("content-disposition", "attachment;filename=\"" + fileName + "\"");
                //context.Response.AddHeader("content-disposition", "attachment;filename=\"" + HttpUtility.UrlEncode(fileName) + "\"");
                context.Response.ContentType = "application/octet-stream";
                context.Response.AddHeader("cache-control", "no-store, no-cache");
                context.Response.ExpiresAbsolute = DateTime.Now.Subtract(new TimeSpan(1, 0, 0, 0));
                context.Response.Expires = -1;
                context.Response.AddHeader("Connection", "Keep-Alive");
                context.Response.AddHeader("accept-ranges", "bytes");

                //request is for byte range
                if (!string.IsNullOrEmpty(context.Request.Headers["Range"]) && context.Request.Headers["Range"].Trim().Length > 0)
                {
                    string range = context.Request.Headers["Range"];
                    range = range.Replace(" ", "");
                    string[] parts = range.Split(new char[] { '=', '-' });
                    long contentLength = long.Parse(parts[2]) - long.Parse(parts[1]);
                    context.Response.AddHeader("content-length", contentLength.ToString());
                    string contentRange = string.Format("bytes {0}-{1}/{2}", parts[1], parts[2], fileLength.ToString());
                    context.Response.AddHeader("content-range", contentRange);
                    context.Response.AddHeader("last-modified", DateTime.Now.ToString("ddd, dd MMM yyyy hh:mm:ss") + " GMT");
                    byte[] bytes = Encoding.ASCII.GetBytes(string.Format("{0} : {1}", fi.Name, fi.LastAccessTimeUtc));
                    string eTag = Convert.ToBase64String(MD5CryptoServiceProvider.Create().ComputeHash(bytes));
                    context.Response.AddHeader("ETag", string.Format("\"{0}\"", eTag));
                    context.Response.StatusCode = 206;  //partial content

                    context.Response.TransmitFile(file, long.Parse(parts[1]), contentLength);
                    context.Response.Flush();
                }
                else  //request is not for byte range
                {
                    context.Response.AddHeader("content-length", fileLength.ToString());

                    if (chunkSize <= 0)
                    {
                        context.Response.TransmitFile(file);
                        context.Response.Flush();
                    }
                    else
                    {
                        long index = 0;
                        while (true)
                        {
                            if (index >= fileLength)
                                break;
                            if (index < fileLength && index + chunkSize > fileLength)
                            {
                                context.Response.TransmitFile(file, index, fileLength - index);
                                context.Response.Flush();
                                break;
                            }

                            context.Response.TransmitFile(file, index, chunkSize);
                            context.Response.Flush();
                            index += chunkSize;
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                context.Response.ClearHeaders();
                context.Response.ClearContent();
                context.Response.Clear();

                context.Response.StatusCode = 500;
                context.Response.Write("Download Error:  " + ex.GetBaseException().Message);
                context.Response.Flush();
                //throw ex;
            }
            finally
            {
                context.Response.End();
                //context.ApplicationInstance.CompleteRequest();
            }
        }

        public bool IsReusable
        {
            get { return false; }
        }
    }

    
    
    ////////////////////////////////////////////////////
    // This custom Basic Authentication module is designed to replace the default built-in 'BasicAuthenticationModule'
    // and use the SqlMembershipProvider (AspNetSqlMembershipProvider with the aspnetdb database) for user security
    // instead of Windows User accounts. When the default Basic Authentication is enabled on the virtual directory, it
    // requires the User to be a Windows User account and verifies the login against the Windows User accounts,
    // and it also requries Forms Authentication to be enabled. If we replace the default built-in 'BasicAuthenticationModule'
    // with our custom Basic Authentication module, Windows User accounts are not required, Forms Authentication does not need
    // to be enabled, and the user is verified against the aspnetdb only.
    //
    // Make sure the aspnetdb database (AspNetSqlMembershipProvider database) is installed on the server
    // by running:  aspnet_regsql.exe -A m
    // (aspnet_regsql.exe is located in \Microsoft.NET\Framework\v4.0)
    //
    // Must make sure database connection string for 'LocalSqlServer' for 'AspNetSqlMembershipProvider' in IIS is pointing at the
    // aspnetdb database that was installed by the above. If use Integrated Security in connection string, we must give the 'NT AUTHORITY\NETWORK SERVICE'
    // account rights to aspnetdb database. If use user name and password in connection string we can use sa account.
    // How to set connection string:
    // In our virtual directory click '.NETUsers' icon in middle of IIS Manager
    //    Click 'Providers' link on right side of IIS Manager
    //       Click 'AspNetSqlMembershipProvider' entry
    //          Click 'ConnectionStrings' link on right side in IIS Manager
    //             Click 'LocalSqlServer' entry
    //                Click 'Edit' link on right side in IIS Manager
    //                   Fill out 'EditConnectionString' dialog
    //
    // Add users you want to have access to web site.
    // To add user:
    //    Click '.NETUsers' icon in middle of IIS Manager
    //       Click 'Add' link on right side in IIS Manager
    //          Fill out 'Add .Net User' dialog
    //
    // Can set the 'Authentication' mode in IIS two different ways to make this work:
    //
    // 1a) Must 'Unlock' BasicAuthenticationModule at server level in IIS so we can remove and replace it with this
    // custom Basic Authentication module at the lower level in the DownloadPortal virtual directory.
    // To 'Unlock':
    //    Click on main server icon in left tree in IIS Manager
    //       Click on 'Modules' icon in middle section in IIS Manager
    //          Click on 'BasicAuthenticationModule' icon in middle section in IIS Manager
    //             Click 'Unlock' link on right side in IIS Manager
    //
    // 1b) Web.config on IIS 7.0:
    //    <system.webServer>
    //      <modules>
    //        <remove name="BasicAuthenticationModule" />
    //        <add name="CustomBasicAuthenticationModule" preCondition="managedHandler" type="DownloadHandlers.CustomBasicAuthenticationModule" />
    //      </modules>
    //    </system.webServer>
    //
    // 1c) 'Disable' ALL 'Authentication' modes in IIS Manager, OR ONLY 'Enable' 'Basic Authentication' in IIS Manager, both will work.
    //     DO NOT 'Enable' 'Anonymous Authentication' because it will not work and it will allow users
    //     to access without being prompted for credentials.
    //        Click on Virtual Directory in left tree in IIS Manager
    //           Click 'Authentication' icon in middle of IIS Manager
    //
    // -- or --
    //
    // 2a) Web.config on IIS 7.0:
    //    <system.webServer>
    //      <modules>
    //        <add name="CustomBasicAuthenticationModule" preCondition="managedHandler" type="DownloadHandlers.CustomBasicAuthenticationModule" />
    //      </modules>
    //    </system.webServer>
    //
    // 2b) 'Disable' ALL 'Authentication' modes in IIS Manager.
    //     DO NOT 'Enable' 'Anonymous Authentication' because it will not work and it will allow users to access without being prompted for credentials.
    //     DO NOT 'Enable' 'Basic Authentication' because it will use the built-in 'Basic Authentication' module that will try to
    //     verify login against the Windows User accounts and it will fail because our login is not a Windows User account, it
    //     is a user in the aspnetdb database.
    //        Click on Virtual Directory in left tree in IIS Manager
    //           Click 'Authentication' icon in middle of IIS Manager
    //
    //    
    // Put DownloadHandler.dll in the /bin directory of the virtual directory on IIS.
    //
    // Note: Make sure the IIS virtual directory is using an AppPool that is using .NET Framework 4.0.
    //       Define a new AppPool using .NET Framework 4.0, if no other AppPool exists that is using .NET Framework 4.0.
    //
    //
    // We should make this download web site use https (SSL) because the 'Basic Authentication' user name and password
    // are being passed as clear text in the http request authentication header.
    // To configure IIS to use https (SSL):
    // 1) Create 'Self-Signed' certificate for the entire server using IIS Manager.
    //    Click on main server icon in left tree in IIS Manager
    //       Click on 'Server Certificates' icon in middle section in IIS Manager
    //          Click 'Create Self-Signed Certificate' link on right side in IIS Manager
    //             Enter friendly name for certificate of 'DownloadPortalCertificate', and it should now show in the 'Server Certificates' list.
    //
    // 2) Add a new 'https' binding entry (https and 443) to the web site virtual directory.
    //       Click on Default Web Site in left tree in IIS Manager
    //          Click 'Bindings' link on right side in IIS Manager
    //             Click 'Add' button in 'Site Bindings' dialog
    //                Enter 'https' and '443' and select the 'DownloadPortalCertificate' certificate we made earlier.
    //
    // 3) Make the web site virtual directory 'Require SSL'.
    //       Click on Virtual Directory in left tree in IIS Manager
    //          Click 'SSL Settings' icon in middle of IIS Manager
    //             Click 'Require SSL' checkbox (leave 'Require 128-bit SSL' unchecked), and (leave the 'Client Certificates' as 'Ignore'), and click 'Apply'.
    ////////////////////////////////////////////////////

    public class CustomBasicAuthenticationModule : IHttpModule
    {
        public void Dispose()
        {
        }

        public void Init(HttpApplication context)
        {
            context.AuthenticateRequest += new EventHandler(context_AuthenticateRequest);
            context.EndRequest += new EventHandler(context_EndRequest);
            context.Error += new EventHandler(context_Error);
            context.AuthorizeRequest += new EventHandler(context_AuthorizeRequest);

            
            #region [Optional] automatically redirect http to https
            //This ONLY works if the 'Require SSL' checkbox is NOT CHECKED in 'SSL Settings' icon for this virtual directory in IIS Manager.
            //'BeginRequest' does NOT fire unless the 'Require SSL' checkbox is NOT CHECKED.
            //This can be used to optionally change ALL http calls to this virtual directory automatically into https calls.
            //This helps for times when user accidentally enters http url and it automatically corrects it to https instead of giving user an error,
            //or if forget to 'Require SSL' on virtual directory in IIS it automatically corrects to https instead of giving user content insecurely.
            //context.BeginRequest += new EventHandler(context_BeginRequest);

            //This WORKS even if the 'Require SSL' checkbox IS CHECKED in 'SSL Settings' icon for this virtual directory in IIS Manager.
            //context.PreSendRequestContent += new EventHandler(context_PreSendRequestContent);
            #endregion
        }

        private void context_AuthenticateRequest(object sender, EventArgs e)
        {
            HttpApplication context = (HttpApplication)sender;

            string authorization = context.Request.Headers["Authorization"];
            if (string.IsNullOrEmpty(authorization))
                return;

            string[] loginArgs = authorization.Split(' ');
            byte[] bytes = Convert.FromBase64String(loginArgs[1]);
            string login = Encoding.ASCII.GetString(bytes);
            string[] user = login.Split(':');
            if (user.GetLength(0) == 0)
                return;

            bool valid = Membership.ValidateUser(user[0], user[1]);
            if (valid)
            {
                context.Context.User = new GenericPrincipal(new GenericIdentity(user[0], "Basic Custom"), null);
            }
            else
            {
                context.Response.StatusCode = 401;
                context.Response.StatusDescription = "Access Denied";
                context.Response.Write("401 Access Denied");
                context.CompleteRequest();
            }
        }
        private void context_EndRequest(object sender, EventArgs e)
        {
            HttpApplication context = (HttpApplication)sender;

            if (context.Response.StatusCode == 401)
                context.Response.AddHeader("WWW-Authenticate", "Basic realm=\"Secured Area\"");
        }
        private void context_Error(object sender, EventArgs e)
        {
            HttpApplication context = (HttpApplication)sender;
            
            //can log error to log file
        }
        private void context_AuthorizeRequest(object sender, EventArgs e)
        {
            HttpApplication context = (HttpApplication)sender;
        }


        #region [Optional] automatically redirect http to https
        //This ONLY works if the 'Require SSL' checkbox is NOT CHECKED in 'SSL Settings' icon for this virtual directory in IIS Manager.
        //'BeginRequest' does NOT fire unless the 'Require SSL' checkbox is NOT CHECKED.
        //This can be used to optionally change ALL http calls to this virtual directory automatically into https calls.
        //This helps for times when user accidentally enters http url and it automatically corrects it to https instead of giving user an error,
        //or if forget to 'Require SSL' on virtual directory in IIS it automatically corrects to https instead of giving user content insecurely.
        //private void context_BeginRequest(object sender, EventArgs e)
        //{
        //    HttpApplication context = (HttpApplication)sender;

        //    if (!context.Request.IsSecureConnection)
        //    {
        //        UriBuilder uri = new UriBuilder(context.Request.Url);
        //        uri.Scheme = "https";
        //        uri.Port = 443;
        //        context.Response.Redirect(uri.Uri.ToString());
        //    }
        //}
        //This WORKS even if the 'Require SSL' checkbox IS CHECKED in 'SSL Settings' icon for this virtual directory in IIS Manager.
        //private void context_PreSendRequestContent(object sender, EventArgs e)
        //{
        //    HttpApplication context = (HttpApplication)sender;

        //    if (context.Response.StatusCode == 403)
        //    {
        //        if (!context.Request.IsSecureConnection)
        //        {
        //            UriBuilder uri = new UriBuilder(context.Request.Url);
        //            uri.Scheme = "https";
        //            uri.Port = 443;
        //            context.Response.Redirect(uri.Uri.ToString());
        //        }
        //    }
        //}
        #endregion

    }


}
